package org.dataone.audit;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.dataone.client.D1Node;
import org.dataone.client.D1TypeBuilder;
import org.dataone.client.ObjectFormatCache;
import org.dataone.client.audit.AuditCNode;
import org.dataone.client.audit.AuditMNode;
import org.dataone.client.audit.D1NodeAuditingWrapper;
import org.dataone.configuration.Settings;
import org.dataone.service.exceptions.BaseException;
import org.dataone.service.exceptions.InvalidRequest;
import org.dataone.service.exceptions.InvalidToken;
import org.dataone.service.exceptions.NotAuthorized;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.NotImplemented;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.v1.Checksum;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.Node;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.NodeType;
import org.dataone.service.types.v1.ObjectFormatIdentifier;
import org.dataone.service.types.v1.ObjectInfo;
import org.dataone.service.types.v1.ObjectList;
import org.dataone.service.types.v1.ObjectLocation;
import org.dataone.service.types.v1.ObjectLocationList;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.service.types.v1.util.ChecksumUtil;
import org.dataone.service.types.v1.util.NodelistUtil;
import org.dataone.service.util.D1Url;

public class AuditTools {

	static Logger logger = Logger.getLogger(AuditTools.class.getName());
	LinkedList<ObjectFormatIdentifier> wantedFormatIds;
	
	/**
	 * a method to allow passing in a file or System.in in the place of a baseUrl
	 * @param filePath
	 * @return
	 * @throws IOException
	 */
	private static List<String> getFilelist(String filePath)
	throws IOException 
	{
		BufferedReader reader = null;
		List<String> list = null;
		try {
			if (filePath.equals('-')) {
				reader = new BufferedReader(new InputStreamReader(System.in));
			} else {
				reader = new BufferedReader(new FileReader(new File(filePath)));
			}

			list = new LinkedList<String>();
			String line = reader.readLine();
			while (line != null) {
				list.add(line);
				line = reader.readLine();
			}
		} finally {
			IOUtils.closeQuietly(reader);
		}
		return list;
	}
	
	
	
	
	public static List<String> listMemberNodeURLs(String baseUrl, boolean includeNodeID) 
	throws NotImplemented, ServiceFailure
	{
		AuditCNode cn = new AuditCNode(baseUrl);
		
		Set<Node> mnSet = NodelistUtil.selectNodes(cn.listNodes(), NodeType.MN);
		
		Iterator<Node> it = mnSet.iterator();
		List<String> urlList = new LinkedList<String>();
		while (it.hasNext()) {
			Node n = it.next();
			if (includeNodeID) {
				urlList.add(n.getBaseURL() + "\t" + n.getIdentifier().getValue());
			} else {
				urlList.add(n.getBaseURL());
			}
		}
		return urlList;
		
	}
	
	
	public static List<String> listIdentifiers(String baseUrl, String[] formatTypes, boolean pageRetrieval)
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, IOException  
	{
		List<String> idList = null;
		if (baseUrl.startsWith("http")) {
			D1Node d1Node = new D1NodeAuditingWrapper(baseUrl + "/v1");

			ObjectList ol = null;
			if (pageRetrieval) {
				ol = APITestUtils.pagedListObjects(d1Node, null, null, null, null, null, null);
			} else {
				Settings.getConfiguration().setProperty("D1Client.D1Node.listObjects.timeout", /*milliseconds*/ 3 * 60 * 1000);
				ol = d1Node.listObjects(null, null, null, null, null, 0, 0);
				logger.info(baseUrl + " objectlist total: " + ol.getTotal());
				logger.info("  getting object list in one gulp...");
				ol = d1Node.listObjects(null, null, null, null, null, 0, ol.getTotal());
			}

			Iterator<ObjectInfo> it = ol.getObjectInfoList().iterator();
			idList = new LinkedList<String>();
			Map<ObjectFormatIdentifier,Boolean> isIncludedFormat = new HashMap<ObjectFormatIdentifier, Boolean>();
			while (it.hasNext()) {
				ObjectInfo oi = it.next();
				String type = null;
				if (formatTypes == null || formatTypes.length == 0 || 
						(isIncludedFormat.get(oi.getFormatId()) != null &&
						isIncludedFormat.get(oi.getFormatId()))) 
				{
					idList.add(oi.getIdentifier().getValue());
				
				} else {
					try {
						type = ObjectFormatCache.getInstance().getFormat(oi.getFormatId()).getFormatType();
						// default to not include unless it matches one of the formatTypes
						isIncludedFormat.put(oi.getFormatId(), false);
						for (String t : formatTypes) {
							if (type.equals(t)) {
								// set format to true
								isIncludedFormat.put(oi.getFormatId(),true);
								idList.add(oi.getIdentifier().getValue());
								break;
							}
						}
					} catch (NotFound e) {
						logger.warn("encountered a formatId without a formatType: " + oi.getFormatId().getValue());
						; // formatType is not found, so treat as not of the requested type
					}
				}

				// remove hopefully helps memory management in the face of large lists
				it.remove();
			}
		} else {
			idList = AuditTools.getFilelist(baseUrl);
		}
		return idList;
	}
	
	public static List<String> listObjectInfos(String baseUrl, String[] formatTypes, boolean pageRetrieval, boolean noDates)
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, IOException  
	{
		List<String> idList = null;
		if (baseUrl.startsWith("http")) {
			D1Node d1Node = new D1NodeAuditingWrapper(baseUrl + "/v1");

			ObjectList ol = null;
			if (pageRetrieval) {
				ol = APITestUtils.pagedListObjects(d1Node, null, null, null, null, null, null);
			} else {
				ol = d1Node.listObjects(null, null, null, null, null, 0, 0);
				logger.info(baseUrl + " objectlist total: " + ol.getTotal());
				logger.info("  getting object list in one gulp...");
				ol = d1Node.listObjects(null, null, null, null, null, 0, ol.getTotal());
			}

			Iterator<ObjectInfo> it = ol.getObjectInfoList().iterator();
			idList = new LinkedList<String>();
			Map<ObjectFormatIdentifier,Boolean> isIncludedFormat = new HashMap<ObjectFormatIdentifier, Boolean>();
			while (it.hasNext()) {
				ObjectInfo oi = it.next();
				String type = null;
				if (formatTypes == null || 
					formatTypes.length == 0 || 
					(isIncludedFormat.get(oi.getFormatId()) != null &&
					isIncludedFormat.get(oi.getFormatId()))) 
				{
					if (noDates) {
						idList.add(String.format("%s\t%s\t%s\t%s\t%s",
								oi.getIdentifier().getValue(),
								oi.getSize(),
								oi.getChecksum().getAlgorithm(),
								oi.getChecksum().getValue(),
								oi.getFormatId().getValue()));
					} else {
						idList.add(String.format("%s\t%s\t%s\t%s\t%s\t%s",
								oi.getIdentifier().getValue(),
								oi.getSize(),
								oi.getChecksum().getAlgorithm(),
								oi.getChecksum().getValue(),
								oi.getFormatId().getValue(),
								oi.getDateSysMetadataModified()));
					}
				} else {
					try {
						type = ObjectFormatCache.getInstance().getFormat(oi.getFormatId()).getFormatType();
						// default to not include unless it matches one of the formatTypes
						isIncludedFormat.put(oi.getFormatId(), false);
						for (String t : formatTypes) {
							if (type.equals(t)) {
								// set format to true
								isIncludedFormat.put(oi.getFormatId(),true);
					
								if (noDates) {
									idList.add(String.format("%s\t%s\t%s\t%s\t%s",
											oi.getIdentifier().getValue(),
											oi.getSize(),
											oi.getChecksum().getAlgorithm(),
											oi.getChecksum().getValue(),
											oi.getFormatId().getValue()));
								} else {
									idList.add(String.format("%s\t%s\t%s\t%s\t%s\t%s",
											oi.getIdentifier().getValue(),
											oi.getSize(),
											oi.getChecksum().getAlgorithm(),
											oi.getChecksum().getValue(),
											oi.getFormatId().getValue(),
											oi.getDateSysMetadataModified()));
								}
								
								break;
							}
						}
					} catch (NotFound e) {
						logger.warn("encountered a formatId without a formatType: " + oi.getFormatId().getValue());
						; // formatType is not found, so treat as not of the requested type
					}
				}
			
				// remove hopefully helps memory management in the face of large lists
				it.remove();
			}
		} else {
			idList = AuditTools.getFilelist(baseUrl);
		}
		return idList;
	}
	


	public static List<String> idListAminusB(String referenceNode, String otherNode, String[] formatTypes) 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, IOException 
	{
		List<String> nodeA = listIdentifiers(referenceNode,formatTypes, false);
		List<String> nodeB = listIdentifiers(otherNode, formatTypes, false);

		List<String> reportAnotB = new LinkedList<String>();
		
		logger.debug("comparing list A to list B...");
		for (String idA : nodeA) {
			if (!nodeB.contains(idA)) {
				reportAnotB.add(idA);
			}
		}
		return reportAnotB;
	}
	
	public static List<String> objectInfosAminusB(String referenceNode, String otherNode, 
			String[] formatTypes, boolean ignoreDates) 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, IOException 
	{
		List<String> nodeA = listObjectInfos(referenceNode, formatTypes, false, ignoreDates);
		List<String> nodeB = listObjectInfos(otherNode, formatTypes, false, ignoreDates);

		List<String> reportAnotB = new LinkedList<String>();

		logger.debug("comparing list A to list B...");
		for (String idA : nodeA) {
			if (!nodeB.contains(idA)) {
				reportAnotB.add(idA);
			}
		}
		return reportAnotB;
	}
	
	public static List<String> searchIDs(String baseUrl) 
	throws InvalidToken, ServiceFailure, NotAuthorized, InvalidRequest, NotImplemented 
	{
		AuditCNode cn = new AuditCNode(baseUrl);
		
		D1Url d1url = new D1Url("foo");
		d1url.addNonEmptyParamPair("q", "*:*");
		

		ObjectList ol = APITestUtils.pagedSolrSearch(cn, d1url, null, null);
		
		Iterator<ObjectInfo> it = ol.getObjectInfoList().iterator();
		List<String> idList = new LinkedList<String>();
		while (it.hasNext()) {
			idList.add(it.next().getIdentifier().getValue());
			// remove hopefully helps memory management in the face of large lists
			it.remove();
		}
		return idList;
	}
	
	
	public static List<String> searchObjectInfos(String baseUrl, boolean noDates) 
	throws InvalidToken, ServiceFailure, NotAuthorized, InvalidRequest, NotImplemented 
	{
		AuditCNode cn = new AuditCNode(baseUrl);
		
		D1Url d1url = new D1Url("foo");
		d1url.addNonEmptyParamPair("q", "*:*");
		

		ObjectList ol = APITestUtils.pagedSolrSearch(cn, d1url, null, null);
		
		Iterator<ObjectInfo> it = ol.getObjectInfoList().iterator();
		List<String> idList = new LinkedList<String>();
		while (it.hasNext()) {
			ObjectInfo oi = it.next();
			if (noDates) {
			idList.add(String.format("%s\t%s\t%s\t%s\t%s",
					oi.getIdentifier().getValue(),
					oi.getSize(),
					oi.getChecksum().getAlgorithm(),
					oi.getChecksum().getValue(),
					oi.getFormatId().getValue()));
			} else {
				idList.add(String.format("%s\t%s\t%s\t%s\t%s\t%s",
						oi.getIdentifier().getValue(),
						oi.getSize(),
						oi.getChecksum().getAlgorithm(),
						oi.getChecksum().getValue(),
						oi.getFormatId().getValue(),
						oi.getDateSysMetadataModified()));
			}
			// remove hopefully helps memory management in the face of large lists
			it.remove();
		}
		return idList;
		
		
	}


	private static String getAuthoritativeMNbaseURL(String cnBaseUrl, NodeReference nodeRef) 
	{
		String result = null;
		AuditCNode cn = new AuditCNode(cnBaseUrl);
		try {
			result = cn.lookupNodeBaseUrl(nodeRef);
		} catch (ServiceFailure e) {
			logger.warn("Could not get the authoritativeMN baseURL from " + cnBaseUrl);
			logger.warn(e.getClass().getName() + ": " + e.getDescription());
		} catch (NotImplemented e) {
			logger.warn("Could not get the authoritativeMN baseURL from " + cnBaseUrl);
			logger.warn(e.getClass().getName() + ": " + e.getDescription());
		}
		return result;
	}
	
	/**
	 * Reports the properties of a systemMetadata object, one per line in xpath 
	 * fashion.  the values of each property follow in subsequent columns per line.
	 * 
	 * @param identifier
	 * @param baseUrls
	 * @return
	 */
	public static List<String> compareSystemMetadata(String identifier, String[] baseUrls, boolean includeSourceMN) {
		
		Map<String,Object> sysmetaMap = new HashMap<String,Object>();
		String sourceBaseUrl = null;
		for (int i = 0; i <= baseUrls.length; i++) {
			String burl = (i < baseUrls.length) ? baseUrls[i] : sourceBaseUrl;
			if (burl == null)
				continue;  
			
			D1Node d1Node = new D1NodeAuditingWrapper(burl + "/v1");
			try {
				SystemMetadata smd = d1Node.getSystemMetadata(null, D1TypeBuilder.buildIdentifier(identifier));
				if (includeSourceMN && sourceBaseUrl == null) {
					NodeReference sourceNodeID = smd.getAuthoritativeMemberNode();
					sourceBaseUrl = getAuthoritativeMNbaseURL(burl,sourceNodeID);
					logger.info("found authoritative node: " + sourceNodeID.getValue() + " " + sourceBaseUrl);
				}
				sysmetaMap.put(burl, smd);				
			}
			catch (Exception e) {
				logger.warn(d1Node.getLatestRequestUrl() + ":: " + e.getClass().getSimpleName() + ": " + e.getMessage());
//				sysmetaMap.put(burl, e.getClass().getSimpleName() + ": " + e.getMessage());
			}
		}
		return compareD1ServiceType(sysmetaMap,true);
	}

	
	/**
	 * Reports the properties of a systemMetadata object, one per line in xpath 
	 * fashion.  the values of each property follow in subsequent columns per line.
	 * 
	 * @param sysmeta1
	 * @param sysmeta2
	 * @return - the report as a List<String> of differences
	 */
	public static List<String> compareSystemMetadata(SystemMetadata sysmeta1, SystemMetadata sysmeta2) {
		
		LinkedHashMap<String,Object> labeledSysmeta = new LinkedHashMap<String,Object>();

		labeledSysmeta.put("sysmeta 1", sysmeta1);
		labeledSysmeta.put("sysmeta 2", sysmeta2);
		
		return compareD1ServiceType(labeledSysmeta,true);
	}
	
	/**
	 * This method will compare any number of dataone service-type objects of 
	 * the same type to each other and return a report of every field where 
	 * there is a difference. Multi-valued fields are sorted by default to minimize 
	 * reporting of meaningless differences, but with the possible side-effect 
	 * of shifting elements out of alignment and reporting more rows than 
	 * differences when there are differences.
	 * 
	 * @param d1ServiceTypeMap - a map of dataone service-type objects
	 * @param sortArrays - Boolean: if set to false, will not sort the arrays prior
	 *        to comparison.
	 * @return - a table of fields where there are differences
	 */
    public static List<String> compareD1ServiceType(Map<String,Object> d1ServiceTypeMap, Boolean sortArrays) {
    	
    	if (sortArrays == null) {
    		sortArrays = true;
    	}
    	
    	List<String> report = new LinkedList<String>();
		Map<String,Object> d1TypePropMap = new HashMap<String,Object>();
		Set<String> propKeys = new TreeSet<String>();
		
		for (String label : d1ServiceTypeMap.keySet()) {
			Map<String,String> smdProps = APITestUtils.getD1SubtypesListing("", d1ServiceTypeMap.get(label), sortArrays);
			propKeys.addAll(smdProps.keySet());
			d1TypePropMap.put(label, smdProps);
		}
    	
		
		Set<String> nodeKeys = d1TypePropMap.keySet();
		StringBuffer sb = new StringBuffer();
		sb.append("property field");
		for (String key: nodeKeys) {
			sb.append("\t" + key);
		}
		report.add(sb.toString());
		
		for (String prop: propKeys) {
			sb = new StringBuffer();
			sb.append(prop);
			boolean difference = false;
			String firstValue = null;
			for (String nodeKey : nodeKeys) {
				sb.append("\t");
				if (d1TypePropMap.get(nodeKey) instanceof LinkedHashMap) {
					String value = ((LinkedHashMap<String,String>)d1TypePropMap.get(nodeKey)).get(prop);
					sb.append(value);
					if (firstValue == null) {
						firstValue = value;
					} else if ( !firstValue.equals(value) ) {
						difference = true;
					}
				} else {
					sb.append(d1TypePropMap.get(nodeKey));
					difference = true;
				}
			}
			if (difference) {
				report.add(sb.toString());
			}
		}
		if (report.size() == 1) {
			report.clear();
			report.add("OK");
		}
		return report;
	}	
	
	
	public static List<String> verifyResolve(String identifier, String cnBaseUrl) 
	throws InvalidToken, ServiceFailure, NotAuthorized, NotFound, NotImplemented 
	{
		AuditCNode cn = new AuditCNode(cnBaseUrl);
		boolean resolved = true;
		Identifier pid = D1TypeBuilder.buildIdentifier(identifier);
		ObjectLocationList oll = cn.resolve(pid);
		List<String> report = new LinkedList<String>();
		for (ObjectLocation ol : oll.getObjectLocationList()) {
			AuditMNode mn = new AuditMNode(ol.getBaseURL());
			try {
				mn.get(pid).close();  // obtain an inputstream for the data, but immediately close it.
				report.add(ol.getBaseURL()+ ": resolved");
			} catch (NotAuthorized be) {
				report.add(ol.getBaseURL() + ": resolved");
			} catch (Exception e) {
				report.add(ol.getBaseURL() + ": " + e.getClass().getSimpleName() +
						": " + e.getMessage());
				resolved = false;
			}
		}
		if (resolved) {
			report.clear();
			report.add("resolve OK");
		} 
		return report;
	}
	
	
	
	/**
	 * verify that the checksum returned from each listed node is the same.
	 * Can be MN or CNs.
	 * @param identifier
	 * @param baseUrls
	 * @param includeSourceMN
	 * @return
	 */
	public static List<String> verifyChecksum(String identifier, String[] baseUrls, boolean includeSourceMN, boolean recalculate) {
		
		List<String> report = new LinkedList<String>();
		Map<String,Object> checksumMap = new HashMap<String,Object>();
		Set<String> propKeys = new TreeSet<String>();
		String sourceBaseUrl = null;
		
		
		// sources for checksums are the passed in urls plus possibly the authoritativeMN 
		for (int i = 0; i < baseUrls.length + 1; i++) {
			String burl = (i < baseUrls.length) ? baseUrls[i] : sourceBaseUrl;
			if (burl == null)
				continue;  
			logger.info("getting checksum from: " + burl);
			D1Node d1Node = new D1NodeAuditingWrapper(burl + "/v1");
			try {
				SystemMetadata smd =  d1Node.getSystemMetadata(D1TypeBuilder.buildIdentifier(identifier));
				if (includeSourceMN && sourceBaseUrl == null) {
					NodeReference sourceNodeID = smd.getAuthoritativeMemberNode();
					sourceBaseUrl = getAuthoritativeMNbaseURL(burl,sourceNodeID);
					logger.info("found authoritative node: " + sourceNodeID.getValue() + " " + sourceBaseUrl);
				}
				Checksum cs = d1Node.getChecksum(null,D1TypeBuilder.buildIdentifier(identifier), null);
				
				Map<String,String> smdProps = APITestUtils.getD1SubtypesListing(cs);
				propKeys.addAll(smdProps.keySet());
				checksumMap.put(burl, smdProps);
				
				if (i == baseUrls.length && recalculate) {
					logger.info("recalculating checksum from authoritative MN");
					AuditMNode mn = new AuditMNode(burl);
					InputStream is = mn.get(D1TypeBuilder.buildIdentifier(identifier));
					cs = ChecksumUtil.checksum(is, cs.getAlgorithm());
					smdProps = APITestUtils.getD1SubtypesListing(cs);
					propKeys.addAll(smdProps.keySet());
					checksumMap.put("recalculated", smdProps);
				}	
			}
			catch (Exception e) {
				logger.warn(d1Node.getLatestRequestUrl() + ":: " + e.getClass().getSimpleName() + ": " + e.getMessage());
				checksumMap.put(burl, e.getClass().getSimpleName() + ": " + e.getMessage());
			}
		}
		
		Set<String> nodeKeys = checksumMap.keySet();
		StringBuffer sb = new StringBuffer();
		sb.append("checksum-field");
		for (String key: nodeKeys) {
			sb.append("\t" + key);
		}
		report.add(sb.toString());
		
		for (String prop: propKeys) {
			sb = new StringBuffer();
			sb.append(prop);
			boolean difference = false;
			String firstValue = null;
			for (String nodeKey : nodeKeys) {
				sb.append("\t");
				
				if (checksumMap.get(nodeKey) instanceof LinkedHashMap) {
					String value = ((LinkedHashMap<String,String>)checksumMap.get(nodeKey)).get(prop);
					sb.append(value);
					if (firstValue == null) {
						firstValue = value;
					} else if ( !firstValue.equals(value) ) {
						difference = true;
					}
				} else {
					sb.append(checksumMap.get(nodeKey));
					difference = true;
				}
			}
			if (difference) {
				report.add(sb.toString());
			}
		}
		if (report.size() == 1) {
			report.clear();
			report.add("OK");
		}
		return report;
	}
	
	public static List<String> compareCnObjectLists(String cnBaseUrl, boolean listObjectInfo)
	{
		boolean pass = true;
		AuditCNode rrCn = null;
		
		HashMap<String,Set<String>> olMap = new HashMap<String,Set<String>>();
		Set<String> superSet = new HashSet<String>();
		
		try {
			rrCn = new AuditCNode(cnBaseUrl);
			
			Set<Node> cnSet = NodelistUtil.selectNodes(rrCn.listNodes(),NodeType.CN);
			
			Date now = new Date();
//			Date toDate = new Date(now.getTime() - 10 * 60 * 1000);

			logger.info("date: " + now);
			
			for (Node n : cnSet) {
				String url = n.getBaseURL();
				// skip the round-robin cn
				if (url.equals(cnBaseUrl))
					continue;
				logger.info(url + ": doing listObjects()");
				try {				
					List<String> idList =  listObjectInfo ? 
							AuditTools.listObjectInfos(url, null, false, false) : 
								AuditTools.listIdentifiers(url, null, false);
					
					logger.info(url + ": adding list to set");
					Set<String> oiSet = listToSet(idList);
					olMap.put(url, oiSet);
					logger.info(url + ": adding set to superset");
					superSet.addAll(oiSet);
					logger.info(String.format("cn: %s   ol size: %d",url, olMap.get(url).size()));
				} 
				catch (BaseException be) {
					pass = false;
					logger.warn("problem getting an ObjectList from the cn. " + be.getClass() + ": " + be.getDescription(), be);
				}
				catch (IOException ioe) {
					pass = false;
					logger.warn("problem getting an List of Identifiers from the file '" + url + "'. " + ioe.getMessage(), ioe);
				}
			}
		}
		catch (BaseException be) {
			logger.warn("problem getting an ObjectList from the cn. " + be.getClass() + ": " + be.getDescription(), be);
		}

		// start reporting the findings
		List<String> report = new LinkedList<String>();

		// create the key to nodes
		Object[] urls = olMap.keySet().toArray();
		report.add("Key to nodes");
		for (int i=0; i<urls.length; i++) {
			report.add((char) (i+65) + " = " + urls[i]);
		}
		report.add("");

		// iterate over the superset and determine which nodes they are on
		Iterator<String> it = superSet.iterator();
		TreeMap<String, String> membershipMap = new TreeMap<String, String>();
		while (it.hasNext()) {
			String coi = it.next();

			String membership = "";
			for (int i=0; i<urls.length; i++) {
				char c = olMap.get(urls[i]).contains(coi) ?  (char) (i+65) : '-';
				membership += c;
			}
			if (membership.contains("-")) {
				membershipMap.put(coi.toString(),membership);
			}
			membership = "";
			it.remove();
		}


		it = membershipMap.keySet().iterator();
		while (it.hasNext()) {
			pass = false;
			String key = it.next();
			report.add(membershipMap.get(key) + "\t" + key);
		}
		if (!pass) {
			report.add("differences were found, or CN was down");
		}
		report.add("difference count = " + membershipMap.keySet().size());

		return report;
	}

	
	/**
	 * a method to convert a list of strings to a set of strings
	 * in a way that does not increase memory requirements unnecessarily 
	 */
	private static Set<String> listToSet(List<String> list) {
		Set<String> set = new HashSet<String>();
		Iterator<String> it = list.iterator();
		while (it.hasNext()) {
			set.add(it.next());
			it.remove();
		}
		return set;
	}
	
}
